/*
 * Copyright (C) 2011 Technische Universitaet Muenchen
 * This file is part of the Sierpi project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpi
 */
/*
 * CGenericTreeNode_tbb.hpp
 *
 *  Created on: Sep 25, 2011
 *      Author: Martin Schreiber (martin.schreiber@in.tum.de)
 */


#ifndef CGENERICTREENODE_TBB_HPP_
#define CGENERICTREENODE_TBB_HPP_

/**
 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * !!! THIS FILE IS INCLUDED DIRECTLY TO THE CGenericTreeNode Class !!!
 * !!! Therefore the following variables are members of the class   !!!
 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 */

#if CONFIG_TBB_TASK_AFFINITIES
	tbb::task::affinity_id	affinity_id;

inline void specializedConstructorMethod()
{
		affinity_id = 0;
}

#else

inline void specializedConstructorMethod()
{
}

#endif

/****************************************************************************
 * GENERIC TREE NODE (NO REDUCE)
 */

/*
 * parallel
 */
template <
	typename CLambdaFun
>
class CTraversalTask_GenericTreeNode_Parallel
	: public tbb::task
{
	CGenericTreeNode_ *this_node;
	CLambdaFun lambda;

public:
	CTraversalTask_GenericTreeNode_Parallel(
			CGenericTreeNode_ *p_this_node,
			CLambdaFun p_lambda
	)	:
		this_node(p_this_node),
		lambda(p_lambda)
	{
#if CONFIG_TBB_TASK_AFFINITIES
		set_affinity(this_node->affinity_id);
#endif
	}


#if CONFIG_TBB_TASK_AFFINITIES
	/**
	 * this method is executed by TBB when a new affinity was set
	 */
	void note_affinity(affinity_id id)
	{
		this_node->affinity_id = id;
	}
#endif

	tbb::task* execute()
	{
		if (this_node->isLeaf())
		{
			lambda(this_node);
			return 0;
		}

		if (this_node->first_child_node != nullptr && this_node->second_child_node != nullptr)
		{
			set_ref_count(3);

			tbb::task &t1 = *new(allocate_child()) CTraversalTask_GenericTreeNode_Parallel<CLambdaFun>(this_node->first_child_node, lambda);
			spawn(t1);

			tbb::task &t2 = *new(allocate_child()) CTraversalTask_GenericTreeNode_Parallel<CLambdaFun>(this_node->second_child_node, lambda);
			spawn_and_wait_for_all(t2);

			return nullptr;
		}


		if (this_node->first_child_node)
		{
			// !!! do the tree-traversal using allocate_child to inherit the tbb tasking information to the class
			// the allocated child seems to be automagically freed
			tbb::task &t1 = *new(allocate_child()) CTraversalTask_GenericTreeNode_Parallel<CLambdaFun>(this_node->first_child_node, lambda);
			t1.execute();
			return nullptr;
		}

//		if (this_node->second_child_node)
//		{
			// !!! do the tree-traversal using allocate_child to inherit the tbb tasking information to the class
			// the allocated child seems to be automagically freed
			tbb::task &t2 = *new(allocate_child()) CTraversalTask_GenericTreeNode_Parallel<CLambdaFun>(this_node->second_child_node, lambda);
			t2.execute();
			return nullptr;
//		}
	}
};


template <typename CLambdaFun>
void traverse_GenericTreeNode_Parallel(
		CLambdaFun p_lambda
)
{
	tbb::task::spawn_root_and_wait((*new(tbb::task::allocate_root()) CTraversalTask_GenericTreeNode_Parallel<CLambdaFun>(this, p_lambda)));
}


/*
 * SERIAL
 */
template <
	typename CLambdaFun
>
class CTraversalTask_GenericTreeNode_Serial
{
	CGenericTreeNode_ *this_node;
	CLambdaFun lambda;

public:
	CTraversalTask_GenericTreeNode_Serial(
			CGenericTreeNode_ *p_this_node,
			CLambdaFun p_lambda
	)	:
		this_node(p_this_node),
		lambda(p_lambda)
	{
	}

	void* execute()
	{
		if (this_node->isLeaf())
		{
			lambda(this_node);
			return 0;
		}

		if (this_node->first_child_node)
			CTraversalTask_GenericTreeNode_Serial<CLambdaFun>(this_node->first_child_node, lambda).execute();

		if (this_node->second_child_node)
			CTraversalTask_GenericTreeNode_Serial<CLambdaFun>(this_node->second_child_node, lambda).execute();

		return 0;
	}
};



template <typename CLambdaFun>
void traverse_GenericTreeNode_Serial(
		CLambdaFun p_lambda
)
{
	CTraversalTask_GenericTreeNode_Serial<CLambdaFun>(this, p_lambda).execute();
}


/*
 * serial/parallel
 */
template <bool parallelProcessing, typename CLambdaFun>
void traverse_GenericTreeNode(
		CLambdaFun p_lambda
)
{
	if (parallelProcessing)
		traverse_GenericTreeNode_Parallel<CLambdaFun>(p_lambda);
	else
		traverse_GenericTreeNode_Serial<CLambdaFun>(p_lambda);
}


/****************************************************************************
 * GENERIC TREE NODE (NO REDUCE) MID AND LEAF NODES
 */

/*
 * parallel
 */
template <
	typename CLambdaFun1,
	typename CLambdaFun2
>
class CTraversalTask_GenericTreeNode_MidAndLeafNodes_Parallel
	: public tbb::task
{
	CGenericTreeNode_ *this_node;
	CLambdaFun1 lambda_leaves;
	CLambdaFun2 lambda_midnodes;

public:
	CTraversalTask_GenericTreeNode_MidAndLeafNodes_Parallel(
			CGenericTreeNode_ *p_this_node,
			CLambdaFun1 i_lambda_leaves,
			CLambdaFun2 i_lambda_midnodes
	)	:
		this_node(p_this_node),
		lambda_leaves(i_lambda_leaves),
		lambda_midnodes(i_lambda_midnodes)
	{
#if CONFIG_TBB_TASK_AFFINITIES
		set_affinity(this_node->affinity_id);
#endif
	}


#if CONFIG_TBB_TASK_AFFINITIES
	/**
	 * this method is executed by TBB when a new affinity was set
	 */
	void note_affinity(affinity_id id)
	{
		this_node->affinity_id = id;
	}
#endif

	tbb::task* execute()
	{
		if (this_node->isLeaf())
		{
			lambda_leaves(this_node);
			return 0;
		}

		if (this_node->first_child_node != nullptr && this_node->second_child_node != nullptr)
		{
			set_ref_count(3);

			tbb::task &t1 = *new( allocate_child()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Parallel<CLambdaFun1, CLambdaFun2>(this_node->first_child_node, lambda_leaves, lambda_midnodes);
			t1.spawn(t1);


			tbb::task &t2 = *new( allocate_child()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Parallel<CLambdaFun1, CLambdaFun2>(this_node->second_child_node, lambda_leaves, lambda_midnodes);

			spawn_and_wait_for_all(t2);

			lambda_midnodes(this_node);

			return nullptr;
		}

		if (this_node->first_child_node)
		{
			// !!! do the tree-traversal using allocate_child to inherit the tbb tasking information to the class
			// the allocated child seems to be automagically freed
			tbb::task &t1 = *new( allocate_child()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Parallel<CLambdaFun1, CLambdaFun2>(this_node->first_child_node, lambda_leaves, lambda_midnodes);
			t1.execute();

			lambda_midnodes(this_node);

			return nullptr;
		}

//		if (this_node->second_child_node)
//		{
			// !!! do the tree-traversal using allocate_child to inherit the tbb tasking information to the class
			// the allocated child seems to be automagically freed
			tbb::task &t2 = *new( allocate_child()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Parallel<CLambdaFun1, CLambdaFun2>(this_node->second_child_node, lambda_leaves, lambda_midnodes);
			t2.execute();

			lambda_midnodes(this_node);

			return nullptr;
//		}
	}
};


template <typename CLambdaFun1, typename CLambdaFun2>
void traverse_GenericTreeNode_MidAndLeafNodes_Parallel(
		CLambdaFun1 i_lambda_leaves,
		CLambdaFun2 i_lambda_midnodes
)
{
	tbb::task::spawn_root_and_wait(*new(tbb::task::allocate_root()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Parallel<CLambdaFun1, CLambdaFun2>(this, i_lambda_leaves, i_lambda_midnodes));
}


/*
 * serial
 */
template <
	typename CLambdaFun1,
	typename CLambdaFun2
>
class CTraversalTask_GenericTreeNode_MidAndLeafNodes_Serial
{
	CGenericTreeNode_ *this_node;
	CLambdaFun1 lambda_leaves;
	CLambdaFun2 lambda_midnodes;

public:
	CTraversalTask_GenericTreeNode_MidAndLeafNodes_Serial(
			CGenericTreeNode_ *p_this_node,
			CLambdaFun1 i_lambda_leaves,
			CLambdaFun2 i_lambda_midnodes
	)	:
		this_node(p_this_node),
		lambda_leaves(i_lambda_leaves),
		lambda_midnodes(i_lambda_midnodes)
	{
	}


	void execute()
	{
		if (this_node->isLeaf())
		{
			lambda_leaves(this_node);
			return;
		}

		if (this_node->first_child_node)
			CTraversalTask_GenericTreeNode_MidAndLeafNodes_Serial<CLambdaFun1, CLambdaFun2>(this_node->first_child_node, lambda_leaves, lambda_midnodes).execute();

		if (this_node->second_child_node)
			CTraversalTask_GenericTreeNode_MidAndLeafNodes_Serial<CLambdaFun1, CLambdaFun2>(this_node->second_child_node, lambda_leaves, lambda_midnodes).execute();

		lambda_midnodes(this_node);
	}
};



template <typename CLambdaFun1, typename CLambdaFun2>
void traverse_GenericTreeNode_MidAndLeafNodes_Serial(
		CLambdaFun1 i_lambda_leaves,
		CLambdaFun2 i_lambda_midnodes
)
{
	CTraversalTask_GenericTreeNode_MidAndLeafNodes_Serial<CLambdaFun1, CLambdaFun2>(this, i_lambda_leaves, i_lambda_midnodes).execute();
}



/*
 * serial/parallel
 */
template <bool parallelProcessing, typename CLambdaFun1, typename CLambdaFun2>
void traverse_GenericTreeNode_MidAndLeafNodes(
		CLambdaFun1 i_lambda_leaves,
		CLambdaFun2 i_lambda_midnodes
)
{
	if (parallelProcessing)
		traverse_GenericTreeNode_MidAndLeafNodes_Parallel(i_lambda_leaves, i_lambda_midnodes);
	else
		traverse_GenericTreeNode_MidAndLeafNodes_Serial(i_lambda_leaves, i_lambda_midnodes);
}



/****************************************************************************
 * GENERIC TREE NODE AND DEPTH FOR ALL MID AND LEAF NODES + DEPTH
 */

/*
 * parallel
 */
template <
	typename CLambdaFun1,
	typename CLambdaFun2
>
class CTraversalTask_GenericTreeNode_MidAndLeafNodes_Depth_Parallel
	: public tbb::task
{
	CGenericTreeNode_ *this_node;
	CLambdaFun1 lambda_leaves;
	CLambdaFun2 lambda_midnodes;
	int genericTreeDepth;

public:
	CTraversalTask_GenericTreeNode_MidAndLeafNodes_Depth_Parallel(
			CGenericTreeNode_ *p_this_node,
			CLambdaFun1 i_lambda_leaves,
			CLambdaFun2 i_lambda_midnodes,
			int p_genericTreeDepth
	)	:
		this_node(p_this_node),
		lambda_leaves(i_lambda_leaves),
		lambda_midnodes(i_lambda_midnodes),
		genericTreeDepth(p_genericTreeDepth)
	{
#if CONFIG_TBB_TASK_AFFINITIES
		set_affinity(this_node->affinity_id);
#endif
	}


#if CONFIG_TBB_TASK_AFFINITIES
	/**
	 * this method is executed by TBB when a new affinity was set
	 */
	void note_affinity(affinity_id id)
	{
		this_node->affinity_id = id;
	}
#endif

	tbb::task* execute()
	{
		if (this_node->isLeaf())
		{
			lambda_leaves(this_node, genericTreeDepth);
			return 0;
		}

		genericTreeDepth++;

		if (this_node->first_child_node != nullptr && this_node->second_child_node != nullptr)
		{
			set_ref_count(3);

			tbb::task &t1 = *new( allocate_child()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Depth_Parallel<CLambdaFun1, CLambdaFun2>(this_node->first_child_node, lambda_leaves, lambda_midnodes, genericTreeDepth);
			spawn(t1);

			tbb::task &t2 = *new( allocate_child()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Depth_Parallel<CLambdaFun1, CLambdaFun2>(this_node->second_child_node, lambda_leaves, lambda_midnodes, genericTreeDepth);
			spawn_and_wait_for_all(t2);

			lambda_midnodes(this_node, genericTreeDepth-1);
			return nullptr;
		}

		if (this_node->first_child_node)
		{
			// !!! do the tree-traversal using allocate_child to inherit the tbb tasking information to the class
			// the allocated child seems to be automagically freed
			tbb::task &t1 = *new( allocate_child()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Depth_Parallel<CLambdaFun1, CLambdaFun2>(this_node->first_child_node, lambda_leaves, lambda_midnodes, genericTreeDepth);
			t1.execute();

			lambda_midnodes(this_node, genericTreeDepth-1);
			return nullptr;
		}

//		if (this_node->second_child_node)
//		{
			// !!! do the tree-traversal using allocate_child to inherit the tbb tasking information to the class
			// the allocated child seems to be automagically freed
			tbb::task &t2 = *new( allocate_child()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Depth_Parallel<CLambdaFun1, CLambdaFun2>(this_node->second_child_node, lambda_leaves, lambda_midnodes, genericTreeDepth);
			t2.execute();

			lambda_midnodes(this_node, genericTreeDepth-1);
			return nullptr;
//		}

		return 0;
	}
};


template <typename CLambdaFun1, typename CLambdaFun2>
void traverse_GenericTreeNode_MidAndLeafNodes_Depth_Parallel(
		CLambdaFun1 i_lambda_leaves,
		CLambdaFun2 i_lambda_midnodes
)
{
	tbb::task::spawn_root_and_wait(*new(tbb::task::allocate_root()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Depth_Parallel<CLambdaFun1, CLambdaFun2>(this, i_lambda_leaves, i_lambda_midnodes, 0));

	if (!this->isLeaf())
		i_lambda_midnodes(this, 0);
}



/*
 * serial
 */
template <
	typename CLambdaFun1,
	typename CLambdaFun2
>
class CTraversalTask_GenericTreeNode_MidAndLeafNodes_Depth_Serial
{
	CGenericTreeNode_ *this_node;
	CLambdaFun1 lambda_leaves;
	CLambdaFun2 lambda_midnodes;
	int genericTreeDepth;

public:
	CTraversalTask_GenericTreeNode_MidAndLeafNodes_Depth_Serial(
			CGenericTreeNode_ *p_this_node,
			CLambdaFun1 i_lambda_leaves,
			CLambdaFun2 i_lambda_midnodes,
			int p_genericTreeDepth
	)	:
		this_node(p_this_node),
		lambda_leaves(i_lambda_leaves),
		lambda_midnodes(i_lambda_midnodes),
		genericTreeDepth(p_genericTreeDepth)
	{
	}

	void execute()
	{
		if (this_node->isLeaf())
		{
			lambda_leaves(this_node, genericTreeDepth);
			return;
		}

		genericTreeDepth++;

		if (this_node->first_child_node)
			CTraversalTask_GenericTreeNode_MidAndLeafNodes_Depth_Serial<CLambdaFun1, CLambdaFun2>(this_node->first_child_node, lambda_leaves, lambda_midnodes, genericTreeDepth).execute();

		if (this_node->second_child_node)
			CTraversalTask_GenericTreeNode_MidAndLeafNodes_Depth_Serial<CLambdaFun1, CLambdaFun2>(this_node->second_child_node, lambda_leaves, lambda_midnodes, genericTreeDepth).execute();

		lambda_midnodes(this_node, genericTreeDepth-1);
	}
};



template <typename CLambdaFun1, typename CLambdaFun2>
void traverse_GenericTreeNode_MidAndLeafNodes_Depth_Serial(
		CLambdaFun1 i_lambda_leaves,
		CLambdaFun2 i_lambda_midnodes
)
{
	CTraversalTask_GenericTreeNode_MidAndLeafNodes_Depth_Serial<CLambdaFun1, CLambdaFun2>(this, i_lambda_leaves, i_lambda_midnodes, 0).execute();

	if (!this->isLeaf())
		i_lambda_midnodes(this, 0);
}



/*
 * serial/parallel
 */
template <bool parallelProcessing, typename CLambdaFun1, typename CLambdaFun2>
void traverse_GenericTreeNode_MidAndLeafNodes_Depth(
		CLambdaFun1 i_lambda_leaves,
		CLambdaFun2 i_lambda_midnodes
)
{
	if (parallelProcessing)
		traverse_GenericTreeNode_MidAndLeafNodes_Depth_Parallel<CLambdaFun1, CLambdaFun2>(i_lambda_leaves, i_lambda_midnodes);
	else
		traverse_GenericTreeNode_MidAndLeafNodes_Depth_Serial<CLambdaFun1, CLambdaFun2>(i_lambda_leaves, i_lambda_midnodes);
}



/****************************************************************************
 * GENERIC TREE NODE (WITH REDUCE)
 ****************************************************************************/
 
/*
 * parallel
 */
template <
	typename CLambdaFun,
	typename TReduceValue
>
class CTraversalTask_GenericTreeNode_Reduce_Parallel	: public tbb::task
{
	CGenericTreeNode_ *this_node;
	CLambdaFun lambda;
	void (*reduceOperator)(const TReduceValue &a, const TReduceValue &b, TReduceValue *o);
	TReduceValue *rootReduceValue;

public:
	TReduceValue reduceValueFirst;
	TReduceValue reduceValueSecond;

	CTraversalTask_GenericTreeNode_Reduce_Parallel(
			CGenericTreeNode_ *i_this_node,
			CLambdaFun i_lambda,
			void (*i_reduceOperator)(const TReduceValue &a, const TReduceValue &b, TReduceValue *o),
			TReduceValue *o_rootReduceValue
	)	:
		this_node(i_this_node),
		lambda(i_lambda),
		reduceOperator(i_reduceOperator),
		rootReduceValue(o_rootReduceValue)
	{
#if CONFIG_TBB_TASK_AFFINITIES
		set_affinity(this_node->affinity_id);
#endif
	}


#if CONFIG_TBB_TASK_AFFINITIES
	/**
	 * this method is executed by TBB when a new affinity was set
	 */
	void note_affinity(affinity_id id)
	{
		this_node->affinity_id = id;
	}
#endif


	/**
	 * TASK
	 */
	task* execute()
	{
		/**
		 * LEAF COMPUTATION
		 */
		if (this_node->isLeaf())
		{
			lambda(this_node, rootReduceValue);

			return nullptr;
		}

		if (this_node->first_child_node != nullptr && this_node->second_child_node != nullptr)
		{
			set_ref_count(3);

			tbb::task &t1 = *new( allocate_child()) CTraversalTask_GenericTreeNode_Reduce_Parallel<CLambdaFun, TReduceValue>(this_node->first_child_node, lambda, reduceOperator, &reduceValueFirst);
			spawn(t1);

			tbb::task &t2 = *new( allocate_child()) CTraversalTask_GenericTreeNode_Reduce_Parallel<CLambdaFun, TReduceValue>(this_node->second_child_node, lambda, reduceOperator, &reduceValueSecond);
			spawn_and_wait_for_all(t2);

			reduceOperator(reduceValueFirst, reduceValueSecond, rootReduceValue);
			return nullptr;
		}

		if (this_node->first_child_node)
		{
			// !!! do the tree-traversal using allocate_child to inherit the tbb tasking information to the class
			// the allocated child seems to be automagically freed
			tbb::task &t1 = *new(allocate_child()) CTraversalTask_GenericTreeNode_Reduce_Parallel<CLambdaFun, TReduceValue>(this_node->first_child_node, lambda, reduceOperator, rootReduceValue);
			t1.execute();

			return nullptr;
		}

//		if (this_node->second_child_node)
//		{
			// !!! do the tree-traversal using allocate_child to inherit the tbb tasking information to the class
			// the allocated child seems to be automagically freed
			tbb::task &t2 = *new(allocate_child()) CTraversalTask_GenericTreeNode_Reduce_Parallel<CLambdaFun, TReduceValue>(this_node->second_child_node, lambda, reduceOperator, rootReduceValue);
			t2.execute();

			return nullptr;
//		}
	}
};

template <typename CLambdaFun, typename TReduceValue>
void traverse_GenericTreeNode_Reduce_Parallel(
		CLambdaFun i_lambda,
		void (*i_reduceOperator)(const TReduceValue &a, const TReduceValue &b, TReduceValue *o),
		TReduceValue *o_reduceOutput
)
{
	tbb::task::spawn_root_and_wait(*new(tbb::task::allocate_root()) CTraversalTask_GenericTreeNode_Reduce_Parallel<CLambdaFun, TReduceValue>(this, i_lambda, i_reduceOperator, o_reduceOutput));
}


/*
 * serial
 */
template <
	typename CLambdaFun,
	typename TReduceValue
>
class CTraversalTask_GenericTreeNode_Reduce_Serial
{
	CGenericTreeNode_ *this_node;
	CLambdaFun lambda;
	void (*reduceOperator)(const TReduceValue &a, const TReduceValue &b, TReduceValue *o);
	TReduceValue *reduceValue;

public:
	TReduceValue reduceValueFirst;
	TReduceValue reduceValueSecond;

	CTraversalTask_GenericTreeNode_Reduce_Serial(
			CGenericTreeNode_ *p_this_node,
			CLambdaFun p_lambda,
			void (*p_reduceOperator)(const TReduceValue &a, const TReduceValue &b, TReduceValue *o),
			TReduceValue *o_reduceValue
	)	:
		this_node(p_this_node),
		lambda(p_lambda),
		reduceOperator(p_reduceOperator),
		reduceValue(o_reduceValue)
	{
	}


	/**
	 * TASK
	 */
	void execute()
	{
		/**
		 * LEAF COMPUTATION
		 */
		if (this_node->isLeaf())
		{
			lambda(this_node, reduceValue);
			return;
		}

		/**
		 * PARALLEL TRAVERSALS
		 */

		if (this_node->first_child_node)
			CTraversalTask_GenericTreeNode_Reduce_Serial<CLambdaFun, TReduceValue>(this_node->first_child_node, lambda, reduceOperator, &reduceValueFirst).execute();

		if (this_node->second_child_node)
			CTraversalTask_GenericTreeNode_Reduce_Serial<CLambdaFun, TReduceValue>(this_node->second_child_node, lambda, reduceOperator, &reduceValueSecond).execute();

		/**
		 * REDUCTION
		 */
		reduceOperator(reduceValueFirst, reduceValueSecond, reduceValue);

		return;
	}
};

template <typename CLambdaFun, typename TReduceValue>
void traverse_GenericTreeNode_Reduce_Serial(
		CLambdaFun i_lambda,
		void (*i_reduceOperator)(const TReduceValue &a, const TReduceValue &b, TReduceValue *o),
		TReduceValue *o_reduceValue
)
{
	CTraversalTask_GenericTreeNode_Reduce_Serial<CLambdaFun, TReduceValue>(this, i_lambda, i_reduceOperator, o_reduceValue).execute();
}


/*
 * serial/parallel
 */
template <bool parallelProcessing, typename CLambdaFun, typename TReduceValue>
void traverse_GenericTreeNode_Reduce(
		CLambdaFun i_lambda,
		void (*i_reduceOperator)(const TReduceValue &i_a, const TReduceValue &i_b, TReduceValue *o_reduceOutput),
		TReduceValue *o_reduceOutput
)
{
	if (parallelProcessing)
		traverse_GenericTreeNode_Reduce_Parallel<CLambdaFun, TReduceValue>(i_lambda, i_reduceOperator, o_reduceOutput);
	else
		traverse_GenericTreeNode_Reduce_Serial<CLambdaFun, TReduceValue>(i_lambda, i_reduceOperator, o_reduceOutput);
}



/****************************************************************************
 * GENERIC TREE NODE MidAndLeafNodes (WITH REDUCE)
 ****************************************************************************/
/*
 * parallel
 */
template <
	typename CLambdaFun1,
	typename CLambdaFun2,
	typename TReduceValue
>
class CTraversalTask_GenericTreeNode_MidAndLeafNodes_Reduce_Parallel	: public tbb::task
{
	CGenericTreeNode_ *this_node;
	CLambdaFun1 lambda_leaves;
	CLambdaFun2 lambda_midnodes;
	void (*reduceOperator)(const TReduceValue &a, const TReduceValue &b, TReduceValue *o);
	TReduceValue *rootReduceValue;

public:
	TReduceValue reduceValueFirst;
	TReduceValue reduceValueSecond;

	CTraversalTask_GenericTreeNode_MidAndLeafNodes_Reduce_Parallel(
			CGenericTreeNode_ *i_this_node,
			CLambdaFun1 i_lambda_leaves,
			CLambdaFun2 i_lambda_midnodes,
			void (*i_reduceOperator)(const TReduceValue &a, const TReduceValue &b, TReduceValue *o),
			TReduceValue *o_rootReduceValue
	)	:
		this_node(i_this_node),
		lambda_leaves(i_lambda_leaves),
		lambda_midnodes(i_lambda_midnodes),
		reduceOperator(i_reduceOperator),
		rootReduceValue(o_rootReduceValue)
	{
#if CONFIG_TBB_TASK_AFFINITIES
		set_affinity(this_node->affinity_id);
#endif
	}


#if CONFIG_TBB_TASK_AFFINITIES
	/**
	 * this method is executed by TBB when a new affinity was set
	 */
	void note_affinity(affinity_id id)
	{
		this_node->affinity_id = id;
	}
#endif


	/**
	 * TASK
	 */
	task* execute()
	{
		/**
		 * LEAF COMPUTATION
		 */
		if (this_node->isLeaf())
		{
			lambda_leaves(this_node, rootReduceValue);
			return nullptr;
		}

		if (this_node->first_child_node != nullptr && this_node->second_child_node != nullptr)
		{
			set_ref_count(3);

			tbb::task &t1 = *new( allocate_child()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Reduce_Parallel<CLambdaFun1, CLambdaFun2, TReduceValue>(this_node->first_child_node, lambda_leaves, lambda_midnodes, reduceOperator, &reduceValueFirst);
			spawn(t1);

			tbb::task &t2 = *new( allocate_child()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Reduce_Parallel<CLambdaFun1, CLambdaFun2, TReduceValue>(this_node->second_child_node, lambda_leaves, lambda_midnodes, reduceOperator, &reduceValueSecond);
			spawn_and_wait_for_all(t2);

			reduceOperator(reduceValueFirst, reduceValueSecond, rootReduceValue);

			lambda_midnodes(this_node, rootReduceValue);

			return nullptr;
		}

		if (this_node->first_child_node)
		{
			// !!! do the tree-traversal using allocate_child to inherit the tbb tasking information to the class
			// the allocated child seems to be automagically freed
			tbb::task &t1 = *new(allocate_child()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Reduce_Parallel<CLambdaFun1, CLambdaFun2, TReduceValue>(this_node->first_child_node, lambda_leaves, lambda_midnodes, reduceOperator, rootReduceValue);
			t1.execute();

			lambda_midnodes(this_node, rootReduceValue);

			return nullptr;
		}

//		if (this_node->second_child_node)
//		{
			// !!! do the tree-traversal using allocate_child to inherit the tbb tasking information to the class
			// the allocated child seems to be automagically freed
			tbb::task &t2 = *new(allocate_child()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Reduce_Parallel<CLambdaFun1, CLambdaFun2, TReduceValue>(this_node->second_child_node, lambda_leaves, lambda_midnodes, reduceOperator, rootReduceValue);
			t2.execute();

			lambda_midnodes(this_node, rootReduceValue);

			return nullptr;
//		}
	}
};

template <
	typename CLambdaFun1,
	typename CLambdaFun2,
	typename TReduceValue
>
void traverse_GenericTreeNode_MidAndLeafNodes_Reduce_Parallel(
		CLambdaFun1 i_lambda_leaves,
		CLambdaFun2 i_lambda_midnodes,
		void (*i_reduceOperator)(const TReduceValue &a, const TReduceValue &b, TReduceValue *o),
		TReduceValue *o_reduceOutput
)
{

	tbb::task::spawn_root_and_wait(*new(tbb::task::allocate_root()) CTraversalTask_GenericTreeNode_MidAndLeafNodes_Reduce_Parallel<CLambdaFun1, CLambdaFun2, TReduceValue>(this, i_lambda_leaves, i_lambda_midnodes, i_reduceOperator, o_reduceOutput));

	if (!this->isLeaf())
		i_lambda_midnodes(this, o_reduceOutput);
}


/*
 * serial
 */
template <
	typename CLambdaFun1,
	typename CLambdaFun2,
	typename TReduceValue
>
class CTraversalTask_GenericTreeNode_MidAndLeafNodes_Reduce_Serial
{
	CGenericTreeNode_ *this_node;
	CLambdaFun1 lambda_leaves;
	CLambdaFun2 lambda_midnodes;
	void (*reduceOperator)(const TReduceValue &a, const TReduceValue &b, TReduceValue *o);
	TReduceValue *rootReduceValue;

public:
	TReduceValue reduceValueFirst;
	TReduceValue reduceValueSecond;

	CTraversalTask_GenericTreeNode_MidAndLeafNodes_Reduce_Serial(
			CGenericTreeNode_ *p_this_node,
			CLambdaFun1 i_lambda_leaves,
			CLambdaFun2 i_lambda_midnodes,
			void (*i_reduceOperator)(const TReduceValue &a, const TReduceValue &b, TReduceValue *o),
			TReduceValue *io_reduceValue
	)	:
		this_node(p_this_node),
		lambda_leaves(i_lambda_leaves),
		lambda_midnodes(i_lambda_midnodes),
		reduceOperator(i_reduceOperator),
		rootReduceValue(io_reduceValue)
	{
	}


	/**
	 * TASK
	 */
	void execute()
	{
		/**
		 * LEAF COMPUTATION
		 */
		if (this_node->isLeaf())
		{
			lambda_leaves(this_node, rootReduceValue);
			return;
		}

		/**
		 * PARALLEL TRAVERSALS
		 */
		if (this_node->first_child_node)
			CTraversalTask_GenericTreeNode_MidAndLeafNodes_Reduce_Serial<CLambdaFun1, CLambdaFun2, TReduceValue>(this_node->first_child_node, lambda_leaves, lambda_midnodes, reduceOperator, &reduceValueFirst).execute();

		if (this_node->second_child_node)
			CTraversalTask_GenericTreeNode_MidAndLeafNodes_Reduce_Serial<CLambdaFun1, CLambdaFun2, TReduceValue>(this_node->second_child_node, lambda_leaves, lambda_midnodes, reduceOperator, &reduceValueSecond).execute();

		/**
		 * REDUCTION
		 */
		reduceOperator(reduceValueFirst, reduceValueSecond, rootReduceValue);

		lambda_midnodes(this_node, rootReduceValue);

		return;
	}
};



template <
	typename CLambdaFun1,
	typename CLambdaFun2,
	typename TReduceValue
>
void traverse_GenericTreeNode_MidAndLeafNodes_Reduce_Serial(
		CLambdaFun1 i_lambda_leaves,
		CLambdaFun2 i_lambda_midnodes,
		void (*i_reduceOperator)(const TReduceValue &a, const TReduceValue &b, TReduceValue *o),
		TReduceValue *o_reduceValue
)
{
	CTraversalTask_GenericTreeNode_MidAndLeafNodes_Reduce_Serial<CLambdaFun1, CLambdaFun2, TReduceValue>(this, i_lambda_leaves, i_lambda_midnodes, i_reduceOperator, o_reduceValue).execute();

	if (!this->isLeaf())
		i_lambda_midnodes(this, o_reduceValue);
}


/*
 * serial/parallel
 */
template <
	bool parallelProcessing,
	typename CLambdaFun1,
	typename CLambdaFun2,
	typename TReduceValue
>
void traverse_GenericTreeNode_MidAndLeafNodes_Reduce(
		CLambdaFun1 i_lambda_leaves,
		CLambdaFun2 i_lambda_midnodes,
		void (*i_reduceOperator)(const TReduceValue &i_a, const TReduceValue &i_b, TReduceValue *o_reduceOutput),
		TReduceValue *o_reduceOutput
)
{
	if (parallelProcessing)
		traverse_GenericTreeNode_MidAndLeafNodes_Reduce_Parallel<CLambdaFun1, CLambdaFun2, TReduceValue>(i_lambda_leaves, i_lambda_midnodes, i_reduceOperator, o_reduceOutput);
	else
		traverse_GenericTreeNode_MidAndLeafNodes_Reduce_Serial<CLambdaFun1, CLambdaFun2, TReduceValue>(i_lambda_leaves, i_lambda_midnodes, i_reduceOperator, o_reduceOutput);
}


#endif
