/*
 * Copyright (C) 2011 Technische Universitaet Muenchen
 * This file is part of the Sierpi project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpi
 *
 *  Created on: Feb 14, 2011
 *      Author: Martin Schreiber (martin.schreiber@in.tum.de)
 */

#ifndef CADAPTIVECONFORMINGGRID_VERTEXCOORDS_NORMALS_DEPTH_DEPTHLIMITERS_ELEMENTDATA_HPP_
#define CADAPTIVECONFORMINGGRID_VERTEXCOORDS_NORMALS_DEPTH_DEPTHLIMITERS_ELEMENTDATA_HPP_

#include <assert.h>
#include "config.h"
#include "libsierpi/triangle/CTriangle_Factory.hpp"
#include "libsierpi/triangle/CTriangle_Normals.hpp"
#include "libsierpi/parallelization/CPartition_SplitJoinInformation.hpp"
#include "CAutomatonTable.hpp"


namespace sierpi
{
namespace travs
{

/*******************************************************
 * !!!!!!!!!! FIRST TRAVERSAL !!!!!!!!!!!
 *******************************************************/

template <typename CKernelClass, typename CSimulationStacks>
class CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_FirstTraversal
{
	typedef typename CSimulationStacks::TElementData TElementData;

public:
	CKernelClass &cKernelClass;

	/**
	 * a few typedefs which could be utilized by the traversator to get more informations
	 */
	typedef typename CKernelClass::TVertexScalar TVertexScalar;
	typedef CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_FirstTraversal<CKernelClass, CSimulationStacks> TThisClass;

private:
	/**
	 * stack and lifo handlers used in this traversator
	 */
	CStack<char> *structure_stack_out;
	CStackReaderTopDown<char> structure_lifo_in;

	/**
	 * lifo iterator and output stack for element data
	 */
	CStackReaderTopDown<TElementData> element_lifo_in;

	/**
	 * stack to store element wise refine/coarsen automata states
	 */
	CStack<unsigned char> *adaptivity_state_flag_stack_out;

public:
	/**
	 * pointer to the corresponding communication stacks
	 */
	CStack<char> *comm_left_edge_stack;
	CStack<char> *comm_right_edge_stack;

	/**
	 * repeat_traversal is set to true, if another refinement iteration has to
	 * be done in order to achieve a conforming grid
	 */
	bool repeat_traversal;


#include "auto/CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_FirstTraversal_autogenerated.hpp"

/**
 * helper functions usage by every parallel traversator
 */
#include "../CTraversatorClassInc_SetupAdaptiveFirstPass.hpp"


	void setup_TraversatorParameters(
			int p_depth_limiter_min,
			int p_depth_limiter_max
	)
	{
		depth_limiter_min = p_depth_limiter_min;
		depth_limiter_max = p_depth_limiter_max;
	}

	CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_FirstTraversal(CKernelClass &p_cKernelClass)	:
				cKernelClass(p_cKernelClass)
	{
		depth_limiter_min = -1;
		depth_limiter_max = -1;
	}

	bool action(CSimulationStacks *cStacks)
	{
		assert(depth_limiter_min != -1);
		assert(depth_limiter_max != -1);

		// INPUT structure stack
		structure_lifo_in.setup(cStacks->structure_stacks.forward);

		// OUTPUT structure stack to output stack in reversed order
		structure_stack_out = &cStacks->structure_stacks.backward;
		structure_stack_out->clear();

		element_lifo_in.setup(cStacks->element_data_stacks.forward);

		comm_left_edge_stack = &cStacks->adaptive_comm_edge_stacks.left;
		comm_right_edge_stack = &cStacks->adaptive_comm_edge_stacks.right;

		adaptivity_state_flag_stack_out = &(cStacks->adaptivity_state_flag_stacks.backward);

		repeat_traversal = false;

		cKernelClass.traversal_pre_hook();

#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
		(this->*sfcRecursiveMethod_Forward)(
				triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
				triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
				triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
				triangleFactory.hypNormal,
				triangleFactory.recursionDepthFirstRecMethod
		);
#else
		(this->*sfcRecursiveMethod_Forward)(
				triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
				triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
				triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
				triangleFactory.recursionDepthFirstRecMethod
		);
#endif

		if (cStacks->structure_stacks.forward.getNumberOfElementsOnStack() == 1)
		{
			/*
			 * check for coarsening request and change it to 'nothing to do'
			 *
			 * the information is encoded in the lower 4 bits
			 */
			if ((adaptivity_state_flag_stack_out->getElementAtIndex(0)&7) == 2)
			{

				adaptivity_state_flag_stack_out->setElementAtIndex(0, 0);
			}
		}

		assert(structure_stack_out->structure_isValidTriangle());
		assert(structure_stack_out->structure_getNumberOfTrianglesInTriangle() == adaptivity_state_flag_stack_out->getNumberOfElementsOnStack());

		if (repeat_traversal)
			return true;

		// 0 => nothing
		// 1 => refine
		// 2 => coarsen
		if (comm_left_edge_stack->query_anyElementNotEqualTo_Value(0))
			return true;
		if (comm_right_edge_stack->query_anyElementNotEqualTo_Value(0))
			return true;

		return false;
	}
};



/*******************************************************
 * !!!!!!!!!! MIDDLE TRAVERSAL !!!!!!!!!!!
 *******************************************************/

template <typename CKernelClass, typename CSimulationStacks>
class CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_MiddleTraversals
{
public:
	typename CSimulationStacks::TElementData TElementData;

	CKernelClass &cKernelClass;

	/**
	 * a few typedefs which could be utilized by the traversator to get more informations
	 */
	typedef CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_MiddleTraversals<CKernelClass,CSimulationStacks> TThisClass;

	/**
	 * stack and lifo handlers used in this traversator
	 */
	CStackReaderTopDown<char> structure_lifo_in;

	/**
	 * stack to store element wise refine/coarsen automata states
	 */
	CStack<unsigned char> *adaptivity_state_flag_stack_in;
	CStack<unsigned char> *adaptivity_state_flag_stack_out;

	/**
	 * pointer to the corresponding communication stacks
	 */
	CStack<char> *comm_left_edge_stack;
	CStack<char> *comm_right_edge_stack;


	/**
	 * repeat_traversal is set to true, if another refinement iteration has to
	 * be done in order to achieve a conforming grid
	 */
	bool repeat_traversal;

#include "auto/CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_MiddleTraversal_autogenerated.hpp"

	/**
	 * helper functions usage by every parallel traversator
	 */
#include "../CTraversatorClassInc_SetupAdaptiveMiddlePasses.hpp"

	CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_MiddleTraversals(CKernelClass &p_cKernelClass)	:
		cKernelClass(p_cKernelClass)
	{
	}


	/**************************************************************************
	 * backward and forward traversals
	 *
	 * \return	true is returned when this action has to be repeated
	 **************************************************************************/
	bool action_Serial(
			CSimulationStacks *cStacks,
			bool i_local_repeat_traversal	///< true, when the traversal has to be repeated
											///< for this sub-partition to create a conforming grid
	)
	{
		structure_lifo_in.setup(cStacks->structure_stacks.backward);

		comm_left_edge_stack = &cStacks->adaptive_comm_edge_stacks.left;
		comm_right_edge_stack = &cStacks->adaptive_comm_edge_stacks.right;

		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.backward;
		adaptivity_state_flag_stack_out = &cStacks->adaptivity_state_flag_stacks.forward;

		repeat_traversal = false;

		// RECURSIVE FUNCTION CALL
		(this->*sfcRecursiveMethod_Backward)();

		assert(comm_left_edge_stack->isEmpty());
		assert(comm_right_edge_stack->isEmpty());
		assert(adaptivity_state_flag_stack_in->isEmpty());

		/**************************************************************************
		 * forward traversal (all edges adjacent to partitions are set to 'new')
		 **************************************************************************/

		// repeat traversal can also be set to true by the backward traversal due to coarsening reasons.
		// therefore it may not be set to false at this point
		//repeat_traversal = false;

		// use default edge stack as adjacent information communication stack
		comm_left_edge_stack = &cStacks->adaptive_comm_edge_stacks.left;
		comm_right_edge_stack = &cStacks->adaptive_comm_edge_stacks.right;
		comm_left_edge_stack->clear();
		comm_right_edge_stack->clear();

		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.forward;
		adaptivity_state_flag_stack_out = &cStacks->adaptivity_state_flag_stacks.backward;

		structure_lifo_in.setup(cStacks->structure_stacks.forward);

//		repeat_traversal = false;

		// RECURSIVE FUNCTION CALL
		(this->*sfcRecursiveMethod_Forward)();

		// final checks
		assert(adaptivity_state_flag_stack_in->isEmpty());


		/**************************************************************************
		 * postprocessing
		 **************************************************************************/

		if (repeat_traversal)
			return true;

		/**
		 * check if a value on the edge is equal to 1.
		 * this means, that there's still a refinement to do since this refinement
		 * request is forwarded to an adjacent partition.
		 *
		 * coarsening is local only. this means, that only a close neighborhood of one
		 * triangle may be coarsened. therefore a check for this is not necessary!
		 */

		if (comm_left_edge_stack->query_anyElementEqualTo_Value(1))
			return true;
		if (comm_right_edge_stack->query_anyElementEqualTo_Value(1))
			return true;

		return false;
	}


	/**************************************************************************
	 * backward and forward traversals
	 *
	 * \return	true is returned when this sub-partition is still in
	 * non-conforming state or if an adjacent sub-partition has to handle some
	 * edge comm data.
	 **************************************************************************/
	bool action_Parallel(
			CSimulationStacks *cStacks,
			bool i_local_repeat_traversal	///< true, when the traversal has to be repeated
											///< for this sub-partition to create a conforming grid
	)
	{
		comm_left_edge_stack = &cStacks->adaptive_comm_exchange_edge_stacks.right;
		comm_right_edge_stack = &cStacks->adaptive_comm_exchange_edge_stacks.left;

		/*
		 * check whether we have to run adaptive traversals
		 *
		 * if the currently processed sub-partition did not request any refinement during
		 * the last traversal and if there's no refinement request from adjacent cells, we
		 * are allowed to skip this traversal.
		 */
#if ADAPTIVE_CONFORMING_SUBPARTITION_SKIPPING_ACTIVE
		if (i_local_repeat_traversal == false)
		{
			/*
			 * run traversal since the current sub-partition is non-conforming
			 */

			/*
			 * check that there is request incoming from adjacent sub-partitions
			 * [ neither refinement (1), nor coarsening (2) ]
			 */
			if (!comm_left_edge_stack->query_anyElementNotEqualTo_Value(0))
			{
				if (!comm_right_edge_stack->query_anyElementNotEqualTo_Value(0))
				{
					/*
					 * check that there is no coarsening request on the local edge comm data
					 */
					if (!cStacks->adaptive_comm_edge_stacks.left.query_anyElementEqualTo_Value(2))
					{
						if (!cStacks->adaptive_comm_edge_stacks.right.query_anyElementEqualTo_Value(2))
						{
							/*
							 * empty exchange stacks since they are read by the backward traversal
							 */
							comm_left_edge_stack->clear();
							comm_right_edge_stack->clear();

							/*
							 * set adaptive edge communication to zero since this stacks are read and analyzed by
							 * the adjacent sub-partitions
							 */
							cStacks->adaptive_comm_edge_stacks.left.setAllElementsEqualToValueToValue(1, 0);
							cStacks->adaptive_comm_edge_stacks.right.setAllElementsEqualToValueToValue(1, 0);
							repeat_traversal = false;

							return false;
						}
					}
				}
			}
		}
#endif

		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.backward;
		adaptivity_state_flag_stack_out = &cStacks->adaptivity_state_flag_stacks.forward;

		structure_lifo_in.setup(cStacks->structure_stacks.backward);

		repeat_traversal = false;

		// RECURSIVE FUNCTION CALL
		(this->*sfcRecursiveMethod_Backward)();

		assert(comm_left_edge_stack->isEmpty());
		assert(comm_right_edge_stack->isEmpty());
		assert(adaptivity_state_flag_stack_in->isEmpty());


		/**************************************************************************
		 * forward traversal (all edges adjacent to partitions are set to 'new')
		 **************************************************************************/

		// repeat traversal can also be set to true by the backward traversal due to coarsening reasons.
		// therefore it may not be set to false at this point
		//repeat_traversal = false;

		structure_lifo_in.setup(cStacks->structure_stacks.forward);
		
		// use default edge stack as adjacent information communication stack
		comm_left_edge_stack = &cStacks->adaptive_comm_edge_stacks.left;
		comm_right_edge_stack = &cStacks->adaptive_comm_edge_stacks.right;
		comm_left_edge_stack->clear();
		comm_right_edge_stack->clear();

		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.forward;
		adaptivity_state_flag_stack_out = &cStacks->adaptivity_state_flag_stacks.backward;


		// RECURSIVE FUNCTION CALL
		(this->*sfcRecursiveMethod_Forward)();

		// final checks
		assert(adaptivity_state_flag_stack_in->isEmpty());


		/**************************************************************************
		 * postprocessing
		 **************************************************************************/

		if (repeat_traversal)
			return true;

		/**
		 * check if a value on the edge is equal to 1.
		 * this means, that there's still a refinement to do since this refinement
		 * request is forwarded to an adjacent partition.
		 *
		 * coarsening is local only. this means, that only a close neighborhood of one
		 * triangle may be coarsened. therefore a check for this is not necessary!
		 */

		if (comm_left_edge_stack->query_anyElementEqualTo_Value(1))
			return true;
		if (comm_right_edge_stack->query_anyElementEqualTo_Value(1))
			return true;

		return false;
	}
};



/*************************************************************************
 * !!!!!!!!!! LAST TRAVERSAL !!!!!!!!!!!
 *
 * this traversal also has to implement the split/join information interface
 *************************************************************************/


template <typename CKernelClass, typename CSimulationStacks>
class CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_LastTraversal
{
public:
	typedef typename CSimulationStacks::TElementData TElementData;
	CKernelClass &cKernelClass;

	/**
	 * a few typedefs which could be utilized by the traversator to get more informations
	 */
	typedef typename CKernelClass::TVertexScalar TVertexScalar;
	typedef CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_LastTraversal<CKernelClass,CSimulationStacks> TThisClass;

	/**
	 * stack and lifo handlers used in this traversator
	 */
	CStack<char> *structure_stack_in;
	CStack<char> *structure_stack_out;
	CStackReaderTopDown<char> structure_lifo_in;

	/**
	 * Fifo iterator and output stack for element data
	 */
	CStackReaderBottomUp<TElementData> element_fifo_in;
	CStack<TElementData> *element_data_stack_out;

	/**
	 * stack to store element wise refine/coarsen automata states
	 */
	CStack<unsigned char> *adaptivity_state_flag_stack_in;
	CStack<unsigned char> *adaptivity_state_flag_stack_out;

	/**
	 * pointer to the corresponding communication stacks
	 */
	CStack<char> *comm_left_edge_stack;
	CStack<char> *comm_right_edge_stack;

	/**
	 * repeat_traversal is set to true, if another refinement iteration has to
	 * be done in order to achieve a conforming grid
	 */
	bool repeat_traversal;

	/**
	 * maximum depth of the previous traversal
	 */
	int maxLeafTraversalDepth;

#include "auto/CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_LastTraversal_autogenerated.hpp"

/**
 * helper functions usage by every parallel traversator
 */
#include "../CTraversatorClassInc_SetupAdaptiveLastPass.hpp"


	/// information which has to be created for the split/join operations supporting parallelization
	CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_LastTraversal(
			CKernelClass &p_cKernelClass
	)	:
			cKernelClass(p_cKernelClass),
			maxLeafTraversalDepth(-1)
	{
	}



	int action_Parallel(
			CSimulationStacks *cStacks,
			CPartition_SplitJoin_EdgeComm_Information &p_splitJoinInformation
	)
	{
#if ADAPTIVE_SUBPARTITION_STACKS
		/*
		 * estimate new size of element data stack.
		 *
		 * this is only an estimation since e. g. coarsening can be forbidden.
		 * therefore we also add an additional offset.
		 *
		 * since the sub-partitions should have cache size, this stack is expected to
		 * be still in cache.
		 */
		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.backward;

		size_t estimated_num_triangles = cStacks->element_data_stacks.forward.getNumberOfElementsOnStack();
		size_t number_of_adaptive_information = adaptivity_state_flag_stack_in->getNumberOfElementsOnStack();


#if ADAPTIVE_SUBPARTITION_AVOID_LAST_TRAVERSAL_IF_NOTHING_CHANGED
		bool at_least_one_refined_or_coarsened_triangle = false;
#endif


		for (size_t i = 0; i < number_of_adaptive_information; i++)
		{
			unsigned char c = adaptivity_state_flag_stack_in->getElementAtIndex(i) & 7;

#if ADAPTIVE_SUBPARTITION_AVOID_LAST_TRAVERSAL_IF_NOTHING_CHANGED
			if (c != 0)
				at_least_one_refined_or_coarsened_triangle = true;
#endif

			if (c == 2)
			{
				/*
				 * special handling of coarsening
				 */
				estimated_num_triangles--;
				i++;
				continue;
			}

			static int deltaTriangles[8] = {0, 0xfffffff, -1, 1, 1, 2, 2, 3};
			estimated_num_triangles += deltaTriangles[c];
		}

		cStacks->changeStackSize_BeforeAdaptivityLastTraversal(estimated_num_triangles);
#endif


		// input/output structure stack
		structure_stack_in = &cStacks->structure_stacks.backward;
		structure_stack_out = &cStacks->structure_stacks.forward;


		cStacks->adaptive_comm_edge_stacks.clear();

		// left and right are exchanged since the first traversal is backwards!
		comm_left_edge_stack = &cStacks->adaptive_comm_edge_stacks.right;
		comm_right_edge_stack = &cStacks->adaptive_comm_edge_stacks.left;

		assert(cStacks->element_data_stacks.backward.isEmpty());


#if ADAPTIVE_SUBPARTITION_AVOID_LAST_TRAVERSAL_IF_NOTHING_CHANGED
		if (!at_least_one_refined_or_coarsened_triangle)
		{
			p_splitJoinInformation.joining_permitted = false;
			p_splitJoinInformation.splitting_permitted = false;

			p_splitJoinInformation.update_edge_comm_required = false;

			maxLeafTraversalDepth = triangleFactory.recursionDepthFirstRecMethod;

			// clear state flag stack
			adaptivity_state_flag_stack_in->clear();

			cKernelClass.traversal_pre_hook();
			cKernelClass.traversal_post_hook();

			return structure_stack_out->structure_getNumberOfTrianglesInTriangle();
		}
#endif


		cKernelClass.traversal_pre_hook();

		element_fifo_in.setup(cStacks->element_data_stacks.forward);
		element_data_stack_out = &cStacks->element_data_stacks.backward;


		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.backward;
		adaptivity_state_flag_stack_out = &cStacks->adaptivity_state_flag_stacks.forward;


		/*
		 * Add offset.
		 *
		 * TODO: Is this really necessary?
		 */
		//estimated_num_triangles += 0;


		// clear output structure stack
		// TODO: is this really necessary?
		structure_stack_out->clear();

		maxLeafTraversalDepth = triangleFactory.recursionDepthFirstRecMethod;

		// RECURSIVE FUNCTIONS

		/***********************************************************
		 * RECURSIVE CALLS - 1ST CHILD
		 ***********************************************************/

		// is this even or odd?
		bool even = (triangleFactory.evenOdd == CTriangle_Enums::EVEN);

		// stack storage place for the hypotenuse
		CStack<char> *hyp_edge_stack = (even ? comm_left_edge_stack : comm_right_edge_stack);

		/***********************************************************
		 * TODO: check if this subtriangle only exists out of a single triangle.
		 * then we have to immediately abort since there's no way to run a split operation on it.
		 * since we can assume to have a basic workload of a few triangles, this shouldn't be a problem
		 */

		/***********************************************************
		 * Sketches given for EVEN triangle factory
		 *
		 * Before split with subtree enumeration
		 *
		 * |\
		 * | 1\
		 * |____\
		 * |    /|\
		 * | 2/3 | 4\
		 * |/____|____\
		 *
		 *
		 * after split
		 * |\
		 * | 1\
		 * |____\
		 * |    /
		 * | 2/   /|\
		 * |/   /3 | 4\ left edge comm stack (identified after traversal of subtree 1) [backward]
		 *    /____|____\
		 *      right edge comm stack [backward]
		 *
		 *
		 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		 * !!!! REMEMBER: THIS TRAVERSAL IS BACKWARDS !!!!!!
		 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		 */

		TVertexScalar mx = (triangleFactory.vertices[0][0] + triangleFactory.vertices[1][0])*(TVertexScalar)0.5;
		TVertexScalar my = (triangleFactory.vertices[0][1] + triangleFactory.vertices[1][1])*(TVertexScalar)0.5;

		/*
		 * check if this subpartition only exists out of a single triangle.
		 */
		if (structure_stack_in->pop() == 0) // pop root level structure stack element
		{
			/**************************************
			 * there's only a single triangle
			 **************************************/
			p_splitJoinInformation.splitting_permitted = false;

			structure_stack_in->push(0);

#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
			int retval =
			(this->*sfcRecursiveMethod_Backward)(
					triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
					triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
					triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
					triangleFactory.hypNormal,
					triangleFactory.recursionDepthFirstRecMethod
				);
#else
			int retval =
			(this->*sfcRecursiveMethod_Backward)(
					triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
					triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
					triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
					triangleFactory.recursionDepthFirstRecMethod
				);
#endif

			if (retval == 0)
			{
				/*
				 * coarsening requested => ignore and fix invalid stacks
				 *
				 * TODO: do we have to fix the comm stacks?
				 * maybe this stacks are not exchanged with adjacent subpartitions
				 */
				for (size_t i = 0; i < comm_left_edge_stack->getNumberOfElementsOnStack(); i++)
					comm_left_edge_stack->setElementAtIndex(i, 0);

				for (size_t i = 0; i < comm_right_edge_stack->getNumberOfElementsOnStack(); i++)
					comm_right_edge_stack->setElementAtIndex(i, 0);

				/*
				 * since the traverser did not store anything on the structure stack, assuming that it is
				 * allowed to coarsen, we have to push an '0' to the structure stack
				 */
				structure_stack_out->push(0);
				element_data_stack_out->push_PtrValue(element_fifo_in.getPreviousDataPtr());
			}

			p_splitJoinInformation.splitting_permitted = false;

			/**
			 * create split/join information for single triangle
			 */

			/**
			 * setup data for pseudo first sub-triangle, but not for second one!
			 */
			p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_right_stack =
					comm_left_edge_stack->getNumberOfElementsOnStack();
			p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_left_stack =
					comm_right_edge_stack->getNumberOfElementsOnStack();
			p_splitJoinInformation.first_triangle.number_of_elements = 1;

			p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_right_stack = 0;
			p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_left_stack = 0;
			p_splitJoinInformation.second_triangle.number_of_elements = 0;

			p_splitJoinInformation.number_of_shared_edge_comm_elements = 0;
		}
		else
		{
			/**************************************
			 * there are at least 2 subtriangles for this subpartition!
			 ***************************************/

			/**************************************
			 * a split operation is not allowed whenever there are only 2 triangles and both request a coarsening
			 **************************************/
#if 0
			/**************************************
			 * a simple solution is to forbid splittings below the minimum recursion depth
			 **************************************/
			p_splitJoinInformation.splitting_permitted = (triangleFactory.recursionDepthFirstRecMethod < triangleFactory.minDepth);
#else
			/**************************************
			 * a more elegant solution is to set the permission to false whenever a coarsening step is taken after this step
			 **************************************/
			p_splitJoinInformation.splitting_permitted = true;
#endif

			/*
			 * backup start stack index to get shared edges
			 */
			int shared_edges_start_counter;

			// pop 2nd level structure stack element
			if (structure_stack_in->pop() == 0)
			{
				/*****************************************************
				 * FIRST sub-subtriangle is also only a single triangle without further refinements
				 *****************************************************/

				// this sub-triangle only has only a recursion level of 1
				maxLeafTraversalDepth = triangleFactory.recursionDepthFirstRecMethod+1;

				// no more leaves => single triangle for the first subtriangle
				// push back the 0 value
				structure_stack_in->push(0);

				if (triangleFactory.evenOdd == CTriangle_Enums::EVEN)
				{
#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
					(this->*sfcRecursiveMethod_Backward_1stChild)(
							triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							mx, my,
							(triangleFactory.hypNormal+5)&7,
							triangleFactory.recursionDepthFirstRecMethod+1
					);
#else
					(this->*sfcRecursiveMethod_Backward_1stChild)(
							triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							mx, my,
							triangleFactory.recursionDepthFirstRecMethod+1
					);
#endif
				}
				else
				{
#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
					(this->*sfcRecursiveMethod_Backward_1stChild)(
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
							mx, my,
							(triangleFactory.hypNormal+3)&7,
							triangleFactory.recursionDepthFirstRecMethod+1
					);
#else
					(this->*sfcRecursiveMethod_Backward_1stChild)(
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
							mx, my,
							triangleFactory.recursionDepthFirstRecMethod+1
					);
#endif
				}

				// remember edge comm data on the left communication stack (we are in backward mode!)
				p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_right_stack =
						comm_left_edge_stack->getNumberOfElementsOnStack();

				// remember edge comm data on the right communication stack (we are in backward mode!)
				p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_left_stack =
						comm_right_edge_stack->getNumberOfElementsOnStack();

				// TODO: [comment]
				p_splitJoinInformation.second_triangle.number_of_elements = element_data_stack_out->getNumberOfElementsOnStack();

				// there can be only a single shared edge element
				p_splitJoinInformation.number_of_shared_edge_comm_elements = 1;

				// start counter to determine edge comm data elements between both elements
				shared_edges_start_counter = hyp_edge_stack->getNumberOfElementsOnStack()-p_splitJoinInformation.number_of_shared_edge_comm_elements;
			}
			else
			{
				/****************************************************
				 * FIRST sub-subtriangle can be further traversed
				 *****************************************************/

				int retval2;
				if (triangleFactory.evenOdd == CTriangle_Enums::EVEN)
				{
					TVertexScalar nmx = (triangleFactory.vertices[1][0] + triangleFactory.vertices[2][0])*0.5;
					TVertexScalar nmy = (triangleFactory.vertices[1][1] + triangleFactory.vertices[2][1])*0.5;

#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
					(this->*sfcRecursiveMethod_Backward_1stChild_1stChild)
						(
							mx, my,
							triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
							nmx, nmy,
							triangleFactory.hypNormal,
							triangleFactory.recursionDepthFirstRecMethod+2
						);
#else
					(this->*sfcRecursiveMethod_Backward_1stChild_1stChild)
						(
							mx, my,
							triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
							nmx, nmy,
							triangleFactory.recursionDepthFirstRecMethod+2
						);
#endif

					shared_edges_start_counter = hyp_edge_stack->getNumberOfElementsOnStack();

#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
					retval2 = (this->*sfcRecursiveMethod_Backward_1stChild_2ndChild)
						(
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							mx, my,
							nmx, nmy,
							(triangleFactory.hypNormal+2)&7,
							triangleFactory.recursionDepthFirstRecMethod+2
						);
#else
					retval2 = (this->*sfcRecursiveMethod_Backward_1stChild_2ndChild)
						(
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							mx, my,
							nmx, nmy,
							triangleFactory.recursionDepthFirstRecMethod+2
						);
#endif
				}
				else
				{
					TVertexScalar nmx = (triangleFactory.vertices[0][0] + triangleFactory.vertices[2][0])*0.5;
					TVertexScalar nmy = (triangleFactory.vertices[0][1] + triangleFactory.vertices[2][1])*0.5;

#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
					(this->*sfcRecursiveMethod_Backward_1stChild_1stChild)
						(
							triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
							mx, my,
							nmx, nmy,
							triangleFactory.hypNormal,
							triangleFactory.recursionDepthFirstRecMethod+2
						);
#else
					(this->*sfcRecursiveMethod_Backward_1stChild_1stChild)
						(
							triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
							mx, my,
							nmx, nmy,
							triangleFactory.recursionDepthFirstRecMethod+2
						);
#endif

					shared_edges_start_counter = hyp_edge_stack->getNumberOfElementsOnStack();


#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
					retval2 = (this->*sfcRecursiveMethod_Backward_1stChild_2ndChild)
						(
							mx, my,
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							nmx, nmy,
							(triangleFactory.hypNormal+6)&7,
							triangleFactory.recursionDepthFirstRecMethod+2
						);
#else
					retval2 = (this->*sfcRecursiveMethod_Backward_1stChild_2ndChild)
						(
							mx, my,
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							nmx, nmy,
							triangleFactory.recursionDepthFirstRecMethod+2
						);
#endif
				}

				// RECURSIVE CALLS for SUBTRIANGLE_2

				/*
				 * when retval2 == 1, no further operation has to be done.
				 * if retval2 == 0, both triangles should be joined.
				 */
				if (retval2 == 1)
				{
					structure_stack_out->push(1);
				}
				else
				{
					/*
					 * if retval2 == 0: both triangles should be joined.
					 */

					/*
					 * forbid split operation since we cannot split a single triangle :-)
					 */
					p_splitJoinInformation.splitting_permitted = false;

					// after coarsening, this sub-triangle only has only a recursion level of 1
					// TODO: is this correct oder should it be -1 instead of +1?
					maxLeafTraversalDepth = triangleFactory.recursionDepthFirstRecMethod+1;

					/*
					 * since retval2 == 0, both subtriangles like to be joined
					 */
					structure_stack_out->push(0);
					TElementData *element = element_data_stack_out->fakePush_returnPtr();

					TElementData *right_element;
					TElementData *left_element;

					if (triangleFactory.evenOdd == CTriangle_Enums::EVEN)
					{
						right_element = element_fifo_in.getPreviousDataPtr();
						left_element = element_fifo_in.getPreviousPreviousDataPtr();

						/*
						 * |\
						 * |  \
						 * | T1 \
						 * |    / \
						 * |  /  T2 \
						 * |/_________\
						 *
						 * backward traversal & EVEN -> create normals for left triangle T2
						 */

						cKernelClass.coarsen(
								triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
								triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
								mx, my,
								CTriangle_Normals::normal_table[triangleFactory.getRightNormalIdx()][0], CTriangle_Normals::normal_table[triangleFactory.getRightNormalIdx()][1],
								CTriangle_Normals::normal_table[triangleFactory.getHypNormalIdx()][1], -CTriangle_Normals::normal_table[triangleFactory.getHypNormalIdx()][0],
								CTriangle_Normals::normal_table[triangleFactory.getHypNormalIdx()][0], CTriangle_Normals::normal_table[triangleFactory.getHypNormalIdx()][1],
								triangleFactory.recursionDepthFirstRecMethod+1,
								element, left_element, right_element
							);
					}
					else
					{
						right_element = element_fifo_in.getPreviousPreviousDataPtr();
						left_element = element_fifo_in.getPreviousDataPtr();

						/* |\
						 * |  \
						 * |    \
						 * |      \
						 * |      /|\
						 * |    /  |  \
						 * |  / T1 | T2 \
						 * |/______|______\
						 *
						 * backward traversal & ODD -> create normals for right triangle T2
						 */

						cKernelClass.coarsen(
								triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
								triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
								mx, my,
								CTriangle_Normals::normal_table[triangleFactory.getLeftNormalIdx()][0], CTriangle_Normals::normal_table[triangleFactory.getLeftNormalIdx()][1],
								CTriangle_Normals::normal_table[triangleFactory.getHypNormalIdx()][0], CTriangle_Normals::normal_table[triangleFactory.getHypNormalIdx()][1],
								-CTriangle_Normals::normal_table[triangleFactory.getHypNormalIdx()][1], CTriangle_Normals::normal_table[triangleFactory.getHypNormalIdx()][0],
								triangleFactory.recursionDepthFirstRecMethod+1,
								element, left_element, right_element
							);
					}
				}

				/***************************************************************************************
				 * BACKUP information about SECOND TRIANGLE:
				 *
				 * remember edge comm data on the left communication stack.
				 * the number has to be stored for the right stack since this traversal goes backwards.
				 */
				p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_right_stack =
						comm_left_edge_stack->getNumberOfElementsOnStack();

				// remember edge comm data on the right communication stack
				p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_left_stack =
						comm_right_edge_stack->getNumberOfElementsOnStack();

				// remember number of elements on the element data stack
				p_splitJoinInformation.second_triangle.number_of_elements = element_data_stack_out->getNumberOfElementsOnStack();


				/***************************************************************************************
				 * BACKUP adaptive refinement/coarsening information on splitting edge
				 *
				 * we have to search the 'number_of_shared_edge_comm_elements' topmost
				 * elements on the left stack for occurrences of refinement or coarsening informations
				 */
				p_splitJoinInformation.number_of_shared_edge_comm_elements = hyp_edge_stack->getNumberOfElementsOnStack() - shared_edges_start_counter;
			}

			bool within_coarsening = false;
			int number_of_shared_edge_comm_elements_delta_elements = 0;

			assert(p_splitJoinInformation.number_of_shared_edge_comm_elements == ((int)hyp_edge_stack->getNumberOfElementsOnStack() - shared_edges_start_counter));

			for (int i = hyp_edge_stack->getNumberOfElementsOnStack()-1; i >= shared_edges_start_counter; i--)
			{
				int a = hyp_edge_stack->getElementAtIndex(i);

				if (a == 0)
					continue;

				if (a == 1)
				{
					number_of_shared_edge_comm_elements_delta_elements++;
					continue;
				}

				assert(a == -1);

				within_coarsening = !within_coarsening;

				if (within_coarsening)
					number_of_shared_edge_comm_elements_delta_elements--;
			}
			assert(within_coarsening == false || p_splitJoinInformation.number_of_shared_edge_comm_elements == 1);



			/***************************************************************************************
			 * RECURSIVE CALLS - 2ND CHILD
			 */
			// pop 2nd level structure stack element

			int retval;
			if (triangleFactory.evenOdd == CTriangle_Enums::EVEN)
			{
#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
				retval = (this->*sfcRecursiveMethod_Backward_2ndChild)(
										triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
										triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
										mx, my,
										(triangleFactory.hypNormal+3)&7,
										triangleFactory.recursionDepthFirstRecMethod+1
									);
#else
				retval = (this->*sfcRecursiveMethod_Backward_2ndChild)(
										triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
										triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
										mx, my,
										triangleFactory.recursionDepthFirstRecMethod+1
									);
#endif
			}
			else
			{
#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
				retval = (this->*sfcRecursiveMethod_Backward_2ndChild)(
										triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
										triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
										mx, my,
										(triangleFactory.hypNormal+5)&7,
										triangleFactory.recursionDepthFirstRecMethod+1
									);
#else
				retval = (this->*sfcRecursiveMethod_Backward_2ndChild)(
										triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
										triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
										mx, my,
										triangleFactory.recursionDepthFirstRecMethod+1
									);
#endif
			}


			// RECURSIVE CALLS for SUBTRIANGLE_2
			if (retval != 0)
			{
				/*
				 * no coarsening operation was requested
				 */
				structure_stack_out->push(retval);
			}
			else
			{
				/*
				 * coarsening operation requested (retval == 0)
				 *
				 * this means that this sub-partition only consists out of 2 triangle which are joined right now!
				 */
				structure_stack_out->push(0);
				TElementData *element_data = element_data_stack_out->fakePush_returnPtr();

				TElementData *right_element;
				TElementData *left_element;

				if (even)
				{
					left_element = element_fifo_in.getPreviousDataPtr();
					right_element = element_fifo_in.getPreviousPreviousDataPtr();

					/*
					 * |\
					 * |  \
					 * | T1 \
					 * |    / \
					 * |  /  T2 \
					 * |/_________\
					 *
					 * backward traversal & EVEN -> create normals for left triangle T2
					 */

					cKernelClass.coarsen(
							triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
							triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							CTriangle_Normals::normal_table[triangleFactory.getHypNormalIdx()][0], CTriangle_Normals::normal_table[triangleFactory.getHypNormalIdx()][1],
							CTriangle_Normals::normal_table[triangleFactory.getRightNormalIdx()][0], CTriangle_Normals::normal_table[triangleFactory.getRightNormalIdx()][1],
							CTriangle_Normals::normal_table[triangleFactory.getLeftNormalIdx()][0], CTriangle_Normals::normal_table[triangleFactory.getLeftNormalIdx()][1],
							triangleFactory.recursionDepthFirstRecMethod,	// this gets to a single triangle -> use recursionDepthFirstRecMethod depth
							element_data,
							left_element,
							right_element
						);
				}
				else
				{
					right_element = element_fifo_in.getPreviousDataPtr();
					left_element = element_fifo_in.getPreviousPreviousDataPtr();

					/*
					 *        /|\
					 *      /  |  \
					 *    / T1 | T2 \
					 *  /______|______\
					 *
					 * backward traversal & ODD -> create normals for right triangle T2
					 */

					cKernelClass.coarsen(
							triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
							triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							CTriangle_Normals::normal_table[triangleFactory.getHypNormalIdx()][0], CTriangle_Normals::normal_table[triangleFactory.getHypNormalIdx()][1],
							CTriangle_Normals::normal_table[triangleFactory.getRightNormalIdx()][0], CTriangle_Normals::normal_table[triangleFactory.getRightNormalIdx()][1],
							CTriangle_Normals::normal_table[triangleFactory.getLeftNormalIdx()][0], CTriangle_Normals::normal_table[triangleFactory.getLeftNormalIdx()][1],
							triangleFactory.recursionDepthFirstRecMethod,	// this gets to a single triangle -> use recursionDepthFirstRecMethod depth
							element_data,
							left_element,
							right_element
						);
				}

				p_splitJoinInformation.splitting_permitted = false;
			}

			p_splitJoinInformation.first_triangle.number_of_elements =
					element_data_stack_out->getNumberOfElementsOnStack()
					- p_splitJoinInformation.second_triangle.number_of_elements;

			if (even)
			{
				p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_left_stack =
						comm_right_edge_stack->getNumberOfElementsOnStack()
						- p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_left_stack;

				p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_right_stack =
						comm_left_edge_stack->getNumberOfElementsOnStack()
						- p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_right_stack
						+ 2*p_splitJoinInformation.number_of_shared_edge_comm_elements;

				p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_right_stack += number_of_shared_edge_comm_elements_delta_elements;
				p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_right_stack += number_of_shared_edge_comm_elements_delta_elements;
			}
			else
			{
				p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_right_stack =
						comm_left_edge_stack->getNumberOfElementsOnStack()
						- p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_right_stack;

				p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_left_stack =
						comm_right_edge_stack->getNumberOfElementsOnStack()
						- p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_left_stack
						+ 2*p_splitJoinInformation.number_of_shared_edge_comm_elements;
						// 2 times, since the shared edge comm elements are already included in number_of_edge_comm_elements_left_stack

				p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_left_stack += number_of_shared_edge_comm_elements_delta_elements;
				p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_left_stack += number_of_shared_edge_comm_elements_delta_elements;
			}


			/***********************************************************
			 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			 * update new shared edge comm elements right now.
			 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			 */
			p_splitJoinInformation.number_of_shared_edge_comm_elements += number_of_shared_edge_comm_elements_delta_elements;
		} // processing of at least 2 triangles in this sub-partition

		// debug
		assert(structure_stack_out->structure_isValidTriangle());
		assert(structure_stack_out->structure_getNumberOfTrianglesInTriangle() == element_data_stack_out->getNumberOfElementsOnStack());

		// clear forward stack and SWAP to make the output stack the forward stack
		cStacks->element_data_stacks.forward.clear();
		cStacks->element_data_stacks.swap();

		cKernelClass.traversal_post_hook();

		p_splitJoinInformation.joining_permitted = true;

		/**
		 * TODO: fix maxLeafTraversalDepth for ignored coarsened triangles
		 */

#if ADAPTIVE_SUBPARTITION_STACKS
		/*
		 * if this assertion fails, add some offset to estimated_num_triangles
		 */
		assert (estimated_num_triangles == cStacks->element_data_stacks.forward.getNumberOfElementsOnStack());

		cStacks->changeStackSize_AfterAdaptivityLastTraversal();
#endif

		p_splitJoinInformation.update_edge_comm_required = true;

		return maxLeafTraversalDepth;
	}


	/**
	 * action without updating the split/join information
	 */
	int action_Serial(CSimulationStacks *cStacks)
	{
#if ADAPTIVE_SUBPARTITION_STACKS
		/*
		 * estimate new size of element data stack.
		 *
		 * this is only an estimation since e. g. coarsening can be forbidden.
		 * therefore we also add an additional offset.
		 *
		 * since the sub-partitions should have cache size, this stack is expected to
		 * be still in cache.
		 */
		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.backward;

		size_t estimated_num_triangles = cStacks->element_data_stacks.forward.getNumberOfElementsOnStack();
		size_t number_of_adaptive_information = adaptivity_state_flag_stack_in->getNumberOfElementsOnStack();

		for (size_t i = 0; i < number_of_adaptive_information; i++)
		{
			unsigned char c = adaptivity_state_flag_stack_in->getElementAtIndex(i) & 7;
			if (c == 2)
			{
				/*
				 * special handling of coarsening
				 */
				estimated_num_triangles--;
				i++;
				continue;
			}

			static int deltaTriangles[8] = {0, 0xfffffff, -1, 1, 1, 2, 2, 3};
			estimated_num_triangles += deltaTriangles[c];
		}

		cStacks->changeStackSize_BeforeAdaptivityLastTraversal(estimated_num_triangles);
#endif
		/**
		 * SETUP LAST TRAVERSAL
		 */

		cStacks->adaptive_comm_edge_stacks.clear();

		element_data_stack_out = &cStacks->element_data_stacks.backward;

		assert(element_data_stack_out->isEmpty());

		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.backward;
		adaptivity_state_flag_stack_out = &cStacks->adaptivity_state_flag_stacks.forward;

		element_fifo_in.setup(cStacks->element_data_stacks.forward);

		// store the output to the backward stack (backward is swapped with the forward stack later on)
		// input/output structure stack
		structure_stack_in = &cStacks->structure_stacks.backward;
		structure_stack_out = &cStacks->structure_stacks.forward;

		// left and right are exchanged since the first traversal is backwards!
		comm_left_edge_stack = &cStacks->adaptive_comm_edge_stacks.right;
		comm_right_edge_stack = &cStacks->adaptive_comm_edge_stacks.left;

		// clear output structure stack
		// TODO: is this really necessary?
		structure_stack_out->clear();

		maxLeafTraversalDepth = 0;

		cKernelClass.traversal_pre_hook();

#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
		(this->*sfcRecursiveMethod_Backward)(
				triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
				triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
				triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
				triangleFactory.hypNormal,
				triangleFactory.recursionDepthFirstRecMethod
			);
#else
		(this->*sfcRecursiveMethod_Backward)(
				triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
				triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
				triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
				triangleFactory.recursionDepthFirstRecMethod
			);
#endif

		cKernelClass.traversal_post_hook();

		// debug
		assert(structure_stack_out->structure_isValidTriangle());
		assert(structure_stack_out->structure_getNumberOfTrianglesInTriangle() == element_data_stack_out->getNumberOfElementsOnStack());

		// clear forward stack and SWAP to make the output stack the forward stack
		cStacks->element_data_stacks.forward.clear();
		cStacks->element_data_stacks.swap();

#if ADAPTIVE_SUBPARTITION_STACKS
		/*
		 * if this assertion fails, add some offset to estimated_num_triangles
		 */
		assert (estimated_num_triangles == cStacks->element_data_stacks.forward.getNumberOfElementsOnStack());

		cStacks->changeStackSize_AfterAdaptivityLastTraversal();
#endif

		return maxLeafTraversalDepth;
	}
};



/**
 * adaptive refinement
 */
template <typename CKernelClass, typename CSimulationStacks>
class CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_ElementData
{
public:
	CKernelClass cKernelClass;
	typedef typename CSimulationStacks::TElementData TElementData;

	CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_FirstTraversal<CKernelClass, CSimulationStacks> firstTraversal;
	CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_MiddleTraversals<CKernelClass, CSimulationStacks> middleTraversals;
	CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Element_LastTraversal<CKernelClass, CSimulationStacks> lastTraversal;

	CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_ElementData()	:
		firstTraversal(cKernelClass),
		middleTraversals(cKernelClass),
		lastTraversal(cKernelClass)
	{
	}



	/**
	 * setup the initial partition traversal for the given factory
	 */
	void setup_RootPartition(
			CTriangle_Factory &i_triangleFactory
	)
	{
		firstTraversal.setup_sfcMethods(i_triangleFactory);
		middleTraversals.setup_sfcMethods(i_triangleFactory);
		lastTraversal.setup_sfcMethods(i_triangleFactory);
	}


	/**
	 * setup the limiters for minimum and maximum depth of curves
	 */
	void setup_RootTraversator(
			int i_depth_limiter_min,
			int i_depth_limiter_max
	)
	{
		firstTraversal.setup_TraversatorParameters(i_depth_limiter_min, i_depth_limiter_max);
	}


	/**
	 * setup the initial partition traversal for the given factory
	 */
	void setup_Partition(
			CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_ElementData<CKernelClass, CSimulationStacks> &i_parent,
			CTriangle_Factory &p_triangleFactory
	)
	{
		firstTraversal.setup_Partition(i_parent.firstTraversal, p_triangleFactory);
		middleTraversals.setup_Partition(i_parent.middleTraversals, p_triangleFactory);
		lastTraversal.setup_Partition(i_parent.lastTraversal, p_triangleFactory);

		// setup kernel class
		cKernelClass.setup_WithKernel(i_parent.cKernelClass);

		firstTraversal.setup_TraversatorParameters(i_parent.firstTraversal.depth_limiter_min, i_parent.firstTraversal.depth_limiter_max);
	}
};

}
}

#endif
